@@ -7,11 +7,9 @@ use warnings;
 use Carp qw(croak cluck);
 use DBI;
 use Class::ReturnValue;
-use Encode;
+use Encode qw();
 
-use vars qw($VERSION @ISA %DBIHandle $PrevHandle $DEBUG %TRANSDEPTH);
-
-$VERSION = '$Version$';
+use vars qw(@ISA %DBIHandle $PrevHandle $DEBUG %TRANSDEPTH);
 
 
 =head1 NAME
@@ -341,31 +339,46 @@ sub dbh {
 
 =head2 Insert $TABLE_NAME @KEY_VALUE_PAIRS
 
-Takes a table name and a set of key-value pairs in an array. splits the key value pairs, constructs an INSERT statement and performs the insert. Returns the row_id of this row.
+Takes a table name and a set of key-value pairs in an array.
+Splits the key value pairs, constructs an INSERT statement
+and performs the insert.
+
+Base class return statement handle object, while DB specific
+subclass should return row id.
 
 =cut
 
 sub Insert {
-  my($self, $table, @pairs) = @_;
-  my(@cols, @vals, @bind);
+    my $self = shift;
+    return $self->SimpleQuery( $self->InsertQueryString(@_) );
+}
 
-  #my %seen; #only the *first* value is used - allows drivers to specify default
-  while ( my $key = shift @pairs ) {
-    my $value = shift @pairs;
-    # next if $seen{$key}++;
-    push @cols, $key;
-    push @vals, '?';
-    push @bind, $value;  
-  }
+=head2 InsertQueryString $TABLE_NAME @KEY_VALUE_PAIRS
 
-  my $QueryString =
-    "INSERT INTO $table (". join(", ", @cols). ") VALUES ".
-    "(". join(", ", @vals). ")";
+Takes a table name and a set of key-value pairs in an array.
+Splits the key value pairs, constructs an INSERT statement
+and returns query string and set of bind values.
 
-    my $sth =  $self->SimpleQuery($QueryString, @bind);
-    return ($sth);
-  }
+This method is more useful for subclassing in DB specific
+handles. L</Insert> method is prefered for end users.
 
+=cut
+
+sub InsertQueryString {
+    my($self, $table, @pairs) = @_;
+    my(@cols, @vals, @bind);
+
+    while ( my $key = shift @pairs ) {
+        push @cols, $key;
+        push @vals, '?';
+        push @bind, shift @pairs;
+    }
+
+    my $QueryString = "INSERT INTO $table";
+    $QueryString .= " (". join(", ", @cols) .")";
+    $QueryString .= " VALUES (". join(", ", @vals). ")";
+    return ($QueryString, @bind);
+}
 
 =head2 UpdateRecordValue 
 
@@ -587,23 +600,48 @@ sub BLOBParams {
 
 
 
-=head2 DatabaseVersion
+=head2 DatabaseVersion [Short => 1]
 
-Returns the database's version. The base implementation uses a "SELECT VERSION"
+Returns the database's version.
+
+If argument C<Short> is true returns short variant, in other
+case returns whatever database handle/driver returns. By default
+returns short version, e.g. '4.1.23' or '8.0-rc4'.
+
+Returns empty string on error or if database couldn't return version.
+
+The base implementation uses a C<SELECT VERSION()>
 
 =cut
 
 sub DatabaseVersion {
     my $self = shift;
+    my %args = ( Short => 1, @_ );
 
-    unless ($self->{'database_version'}) {
-        my $statement  = "SELECT VERSION()";
-        my $sth = $self->SimpleQuery($statement);
-        my @vals = $sth->fetchrow();
-        $self->{'database_version'}= $vals[0];
+    unless ( defined $self->{'database_version'} ) {
+
+        # turn off error handling, store old values to restore later
+        my $re = $self->RaiseError;
+        $self->RaiseError(0);
+        my $pe = $self->PrintError;
+        $self->PrintError(0);
+
+        my $statement = "SELECT VERSION()";
+        my $sth       = $self->SimpleQuery($statement);
+
+        my $ver = '';
+        $ver = ( $sth->fetchrow_arrayref->[0] || '' ) if $sth;
+        $ver =~ /(\d+(?:\.\d+)*(?:-[a-z0-9]+)?)/i;
+        $self->{'database_version'}       = $ver;
+        $self->{'database_version_short'} = $1 || $ver;
+
+        $self->RaiseError($re);
+        $self->PrintError($pe);
     }
-}
 
+    return $self->{'database_version_short'} if $args{'Short'};
+    return $self->{'database_version'};
+}
 
 =head2 CaseSensitive
 
@@ -837,8 +875,8 @@ sub Join {
     my %args = (
         SearchBuilder => undef,
         TYPE          => 'normal',
-        FIELD1        => 'main',
-        ALIAS1        => undef,
+        ALIAS1        => 'main',
+        FIELD1        => undef,
         TABLE2        => undef,
         FIELD2        => undef,
         ALIAS2        => undef,
@@ -846,7 +884,6 @@ sub Join {
         @_
     );
 
-    my $string;
 
     my $alias;
 
@@ -859,10 +896,9 @@ sub Join {
         my @aliases = @{ $args{'SearchBuilder'}->{'aliases'} };
         my @new_aliases;
         foreach my $old_alias (@aliases) {
-            if ( $old_alias =~ /^(.*?) ($args{'ALIAS2'})$/ ) {
+            if ( $old_alias =~ /^(.*?) (\Q$args{'ALIAS2'}\E)$/ ) {
                 $args{'TABLE2'} = $1;
                 $alias = $2;
-
             }
             else {
                 push @new_aliases, $old_alias;
@@ -895,7 +931,8 @@ sub Join {
 
         }
 
-        if ( !$alias || $args{'ALIAS1'} ) {
+        unless ( $alias ) {
+            # XXX: this situation is really bug in the caller!!!
             return ( $self->_NormalJoin(%args) );
         }
         $args{'SearchBuilder'}->{'aliases'} = \@new_aliases;
@@ -906,31 +943,20 @@ sub Join {
 
     }
 
+    my $meta = $args{'SearchBuilder'}->{'left_joins'}{"$alias"} ||= {};
     if ( $args{'TYPE'} =~ /LEFT/i ) {
-
-        $string = " LEFT JOIN " . $args{'TABLE2'} . " $alias ";
-
+        $meta->{'alias_string'} = " LEFT JOIN " . $args{'TABLE2'} . " $alias ";
+        $meta->{'type'} = 'LEFT';
     }
     else {
-
-        $string = " JOIN " . $args{'TABLE2'} . " $alias ";
-
+        $meta->{'alias_string'} = " JOIN " . $args{'TABLE2'} . " $alias ";
+        $meta->{'type'} = 'NORMAL';
     }
+    $meta->{'depends_on'} = $args{'ALIAS1'};
 
-
-    my $criterion;
-    if ($args{'EXPRESSION'}) {
-        $criterion = $args{'EXPRESSION'};
-    } else {
-        $criterion = $args{'ALIAS1'}.".".$args{'FIELD1'};
-    }
-
-    $args{'SearchBuilder'}->{'left_joins'}{"$alias"}{'alias_string'} = $string;
-    $args{'SearchBuilder'}->{'left_joins'}{"$alias"}{'depends_on'}   =
-      $args{'ALIAS1'};
-    $args{'SearchBuilder'}->{'left_joins'}{"$alias"}{'criteria'}
-      { 'criterion' . $args{'SearchBuilder'}->{'criteria_count'}++ } =
-      " $alias.$args{'FIELD2'} = $criterion";
+    my $criterion = $args{'EXPRESSION'} || $args{'ALIAS1'}.".".$args{'FIELD1'};
+    $meta->{'criteria'}{'base_criterion'} =
+        [ { field => "$alias.$args{'FIELD2'}", op => '=', value => $criterion } ];
 
     return ($alias);
 }
@@ -953,11 +979,15 @@ sub _NormalJoin {
 
     if ( $args{'TYPE'} =~ /LEFT/i ) {
         my $alias = $sb->_GetAlias( $args{'TABLE2'} );
-        $sb->{'left_joins'}{"$alias"}{'alias_string'} =
-          " LEFT JOIN $args{'TABLE2'} $alias ";
-
-        $sb->{'left_joins'}{"$alias"}{'criteria'}{'base_criterion'} =
-          " $args{'ALIAS1'}.$args{'FIELD1'} = $alias.$args{'FIELD2'}";
+        my $meta = $sb->{'left_joins'}{"$alias"} ||= {};
+        $meta->{'alias_string'} = " LEFT JOIN $args{'TABLE2'} $alias ";
+        $meta->{'depends_on'}   = $args{'ALIAS1'};
+        $meta->{'type'}         = 'LEFT';
+        $meta->{'criteria'}{'base_criterion'} = [ {
+            field => "$args{'ALIAS1'}.$args{'FIELD1'}",
+            op => '=',
+            value => "$alias.$args{'FIELD2'}",
+        } ];
 
         return ($alias);
     }
@@ -979,50 +1009,206 @@ sub _NormalJoin {
 sub _BuildJoins {
     my $self = shift;
     my $sb   = shift;
-    my %seen_aliases;
 
-    $seen_aliases{'main'} = 1;
+    $self->OptimizeJoins( SearchBuilder => $sb );
 
-   	# We don't want to get tripped up on a dependency on a simple alias. 
-    	foreach my $alias ( @{ $sb->{'aliases'}} ) {
-          if ( $alias =~ /^(.*?)\s+(.*?)$/ ) {
-              $seen_aliases{$2} = 1;
-          }
+    my $join_clause = join " CROSS JOIN ", ($sb->Table ." main"), @{ $sb->{'aliases'} };
+    my %processed = map { /^\S+\s+(\S+)$/; $1 => 1 } @{ $sb->{'aliases'} };
+    $processed{'main'} = 1;
+
+    # get a @list of joins that have not been processed yet, but depend on processed join
+    my $joins = $sb->{'left_joins'};
+    while ( my @list = grep !$processed{ $_ }
+            && $processed{ $joins->{ $_ }{'depends_on'} }, keys %$joins )
+    {
+        foreach my $join ( @list ) {
+            $processed{ $join }++;
+
+            my $meta = $joins->{ $join };
+            my $aggregator = $meta->{'entry_aggregator'} || 'AND';
+
+            $join_clause .= $meta->{'alias_string'} . " ON ";
+            my @tmp = map {
+                    ref($_)?
+                        $_->{'field'} .' '. $_->{'op'} .' '. $_->{'value'}:
+                        $_
+                }
+                map { ('(', @$_, ')', $aggregator) } values %{ $meta->{'criteria'} };
+            pop @tmp;
+            $join_clause .= join ' ', @tmp;
+        }
     }
 
-    my $join_clause = $sb->Table . " main ";
-
-	
-    my @keys = ( keys %{ $sb->{'left_joins'} } );
-    my %seen;
-
-    while ( my $join = shift @keys ) {
-        my $aggregator = $sb->{'left_joins'}{$join}{'entry_aggregator'} 
-            || 'AND';
-
-        if ( ! $sb->{'left_joins'}{$join}{'depends_on'} || $seen_aliases{ $sb->{'left_joins'}{$join}{'depends_on'} } ) {
-            $join_clause = "(" . $join_clause;
-            $join_clause .=
-              $sb->{'left_joins'}{$join}{'alias_string'} . " ON (";
-            $join_clause .=
-              join ( ") $aggregator ( ",
-                values %{ $sb->{'left_joins'}{$join}{'criteria'} } );
-            $join_clause .= ")) ";
-
-            $seen_aliases{$join} = 1;
-        }
-        else {
-            push ( @keys, $join );
-            die "Unsatisfied dependency chain in Joins @keys"
-              if $seen{"@keys"}++;
-        }
-
+    # here we could check if there is recursion in joins by checking that all joins
+    # are processed
+    if ( my @not_processed = grep !$processed{ $_ }, keys %$joins ) {
+        die "Unsatisfied dependency chain in joins @not_processed";
     }
-    return ( join ( ", ", ( $join_clause, @{ $sb->{'aliases'} } ) ) );
+    return $join_clause;
+}
+
+sub OptimizeJoins {
+    my $self = shift;
+    my %args = (SearchBuilder => undef, @_);
+    my $joins = $args{'SearchBuilder'}->{'left_joins'};
+
+    my %processed = map { /^\S+\s+(\S+)$/; $1 => 1 } @{ $args{'SearchBuilder'}->{'aliases'} };
+    $processed{ $_ }++ foreach grep $joins->{ $_ }{'type'} ne 'LEFT', keys %$joins;
+    $processed{'main'}++;
+
+    my @ordered;
+    # get a @list of joins that have not been processed yet, but depend on processed join
+    # if we are talking about forest then we'll get the second level of the forest,
+    # but we should process nodes on this level at the end, so we build FILO ordered list.
+    # finally we'll get ordered list with leafes in the beginning and top most nodes at
+    # the end.
+    while ( my @list = grep !$processed{ $_ }
+            && $processed{ $joins->{ $_ }{'depends_on'} }, keys %$joins )
+    {
+        unshift @ordered, @list;
+        $processed{ $_ }++ foreach @list;
+    }
+
+    foreach my $join ( @ordered ) {
+        next if $self->MayBeNull( SearchBuilder => $args{'SearchBuilder'}, ALIAS => $join );
+
+        $joins->{ $join }{'alias_string'} =~ s/^\s*LEFT\s+/ /;
+        $joins->{ $join }{'type'} = 'NORMAL';
+    }
+
+    # here we could check if there is recursion in joins by checking that all joins
+    # are processed
 
 }
 
+=head2 MayBeNull
 
+Takes a C<SearchBuilder> and C<ALIAS> in a hash and resturns
+true if restrictions of the query allow NULLs in a table joined with
+the ALIAS, otherwise returns false value which means that you can
+use normal join instead of left for the aliased table.
+
+Works only for queries have been built with L<DBIx::SearchBuilder/Join> and
+L<DBIx::SearchBuilder/Limit> methods, for other cases return true value to
+avoid fault optimizations.
+
+=cut
+
+sub MayBeNull {
+    my $self = shift;
+    my %args = (SearchBuilder => undef, ALIAS => undef, @_);
+    # if we have at least one subclause that is not generic then we should get out
+    # of here as we can't parse subclauses
+    return 1 if grep $_ ne 'generic_restrictions', keys %{ $args{'SearchBuilder'}->{'subclauses'} };
+
+    # build full list of generic conditions
+    my @conditions;
+    foreach ( grep @$_, values %{ $args{'SearchBuilder'}->{'restrictions'} } ) {
+        push @conditions, 'AND' if @conditions;
+        push @conditions, '(', @$_, ')';
+    }
+
+    # find tables that depends on this alias and add their join conditions
+    foreach my $join ( values %{ $args{'SearchBuilder'}->{'left_joins'} } ) {
+        # left joins on the left side so later we'll get 1 AND x expression
+        # which equal to x, so we just skip it
+        next if $join->{'type'} eq 'LEFT';
+        next unless $join->{'depends_on'} eq $args{'ALIAS'};
+
+        my @tmp = map { ('(', @$_, ')', $join->{'entry_aggregator'}) } values %{ $join->{'criteria'} };
+        pop @tmp;
+
+        @conditions = ('(', @conditions, ')', 'AND', '(', @tmp ,')');
+
+    }
+    return 1 unless @conditions;
+
+    # replace conditions with boolean result: 1 - allows nulls, 0 - not
+    # all restrictions on that don't act on required alias allow nulls
+    # otherwise only IS NULL operator 
+    foreach ( splice @conditions ) {
+        unless ( ref $_ ) {
+            push @conditions, $_;
+        } elsif ( $_->{'field'} =~ /^\Q$args{'ALIAS'}./ ) {
+            push @conditions, lc $_->{op} eq 'is';
+        } elsif ( $_->{'value'} && $_->{'value'} =~ /^\Q$args{'ALIAS'}./ ) {
+            push @conditions, 0;
+        } else {
+            push @conditions, 1;
+        }
+    }
+
+    # resturns index of closing paren by index of openning paren
+    my $closing_paren = sub {
+        my $i = shift;
+        my $count = 0;
+        for ( ; $i < @conditions; $i++ ) {
+            if ( $conditions[$i] eq '(' ) {
+                $count++;
+            }
+            elsif ( $conditions[$i] eq ')' ) {
+                $count--;
+            }
+            return $i unless $count;
+        }
+        die "lost in parens";
+    };
+
+    # solve boolean expression we have, an answer is our result
+    my @tmp = ();
+    while ( defined ( my $e = shift @conditions ) ) {
+        #warn "@tmp >>>$e<<< @conditions";
+        return $e if !@conditions && !@tmp;
+
+        unless ( $e ) {
+            if ( $conditions[0] eq ')' ) {
+                push @tmp, $e;
+                next;
+            }
+
+            my $aggreg = shift @conditions;
+            if ( $aggreg eq 'OR' ) {
+                # 0 OR x == x
+                next;
+            } elsif ( $aggreg eq 'AND' ) {
+                # 0 AND x == 0
+                my $close_p = $closing_paren->(0);
+                splice @conditions, 0, $close_p + 1, (0);
+            } else {
+                die "lost @tmp >>>$e $aggreg<<< @conditions";
+            }
+        } elsif ( $e eq '1' ) {
+            if ( $conditions[0] eq ')' ) {
+                push @tmp, $e;
+                next;
+            }
+
+            my $aggreg = shift @conditions;
+            if ( $aggreg eq 'OR' ) {
+                # 1 OR x == 1
+                my $close_p = $closing_paren->(0);
+                splice @conditions, 0, $close_p + 1, (1);
+            } elsif ( $aggreg eq 'AND' ) {
+                # 1 AND x == x
+                next;
+            } else {
+                die "lost @tmp >>>$e $aggreg<<< @conditions";
+            }
+        } elsif ( $e eq '(' ) {
+            if ( $conditions[1] eq ')' ) {
+                splice @conditions, 1, 1;
+            } else {
+                push @tmp, $e;
+            }
+        } elsif ( $e eq ')' ) {
+            unshift @conditions, @tmp, $e;
+            @tmp = ();
+        } else {
+            die "lost: @tmp >>>$e<<< @conditions";
+        }
+    }
+    return 1;
+}
 
 =head2 DistinctQuery STATEMENTREF 
 
